#include "opencv2/objdetect/objdetect.hpp"
#include "opencv2/highgui/highgui.hpp"
#include "opencv2/imgproc/imgproc.hpp"
#include <iostream>
#include <stdio.h>
#include "cascadeclassifier.h"

#include <gl/glut.h>
#include <gl/GL.h>
#include "Geometry.h"
#include "Quat.h"


#define M_LEFT 0
#define M_RIGHT 1
#define M_UP 2
#define M_DOWN 3
#define M_STATIC -1

using namespace std;
using namespace cv;

//** Function Headers */
void detectAndDisplay( Mat frame );

//** Global variables */
String face_cascade_name = "haarcascade_frontalface_alt.xml";
String eyes_cascade_name = "haarcascade_eye_tree_eyeglasses.xml";
CascadeClassifier face_cascade;
CascadeClassifier eyes_cascade;
string window_name = "Capture - Face detection";
RNG rng(12345);

Point last_center;
double last_box_size=1.0;

CvCapture* capture;
Mat frame;

//opengl variables
////////////added////////////////
GLfloat dx = 0.f, dy = 0.f;
GLfloat rot_x = 0.f, rot_y = 0.f; 

//////////////////////////////////
float whratio;
int gButton;
int mouseX, mouseY;
int win_width, win_height;
/* rotation quaternion and translation vector for the object */
CQrot       ObjRot(1,0,0,0);
Vector3D    ObjTrans(0,0,0);

/* rotation quaternion and translation vector for the object */
CQrot       EyeRot(1,0,0,0);
Vector3D    EyeTrans(0,0,5);

/* rotation quaternion for the light */
CQrot       LightRot(1,0,0,0);

void cvDetectFaceMotion();
int val_time_callback;
int hor_move, vert_move;

void timer(int p)
{
	cvDetectFaceMotion();
	glutPostRedisplay();       //marks the current window as needing to be redisplayed.
	glutTimerFunc(20, timer,0);
}

void setupGLstate(){

	GLfloat lightOneColor[] = {1, 1, 1, 1};
	GLfloat globalAmb[] = {.1, .1, .1, 1};
	GLfloat lightOnePosition[] = {.0, .2, 1, 0.0};

	glEnable(GL_CULL_FACE);
	glFrontFace(GL_CCW);      
	glEnable(GL_DEPTH_TEST);
	glClearColor(0,0,0,0);
	glShadeModel(GL_SMOOTH);

	glEnable(GL_LIGHT1);
	glEnable(GL_LIGHTING);
	glEnable(GL_NORMALIZE);
	glEnable(GL_COLOR_MATERIAL);

	glLightfv(GL_LIGHT1, GL_DIFFUSE, lightOneColor);
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, globalAmb);
	glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);

	glLightfv(GL_LIGHT1, GL_POSITION, lightOnePosition);
}

void reshapeScene(int width, int height)
{
	win_width=width;
	win_height=height;

	GLsizei mwidth = (width<height*whratio)?width:(height*whratio);

	glViewport(0,0,mwidth, (float)mwidth/whratio);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity(); 

	gluPerspective( 40.0, /* field of view in degrees */
		whratio, /* aspect ratio */
		0.01, /* Z near */ 
		1e3 /* Z far */);
	//gluLookAt(0.0, 0.0, 0.0, -1.0, 0.0, 0.0, 1, 0, 0);
	glMatrixMode(GL_MODELVIEW);

	glutPostRedisplay();
}

void setupObject(void)
{
	double rot[16];
	glTranslated( ObjTrans.x, ObjTrans.y, ObjTrans.z);
	ObjRot.convert( rot );
	glMultMatrixd(( GLdouble *)  rot );	
}

/* the eye is always fixed at world z = +5 */
void setupEye(void)
{
	glLoadIdentity();

	double rot[16];
	CQrot eye = EyeRot;
	eye^(-1);
	eye.convert( rot );

	glTranslated( -EyeTrans.x, -EyeTrans.y, -EyeTrans.z); 
	glMultMatrixd(( GLdouble *)  rot );
}

/* setup light */
void setupLight()
{
	Vector3D position(0,0,1);
	position = LightRot* position;
	GLfloat lightOnePosition[4]={position.x, position.y, position.z, 0};
	glLightfv(GL_LIGHT1, GL_POSITION, lightOnePosition);
}
void renderScene(void) 
{
	//if (){
	//	cvDetectFaceMotion();
	//}


	/* clear frame buffer */
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	//setupLight();
	/* transform from the eye coordinate system to the world system */
	setupEye();

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	/* transform from the world to the ojbect coordinate system */
	setupObject();

	///////Tranlsate and Rotate////////
	glTranslatef(dx, dy, -5.0f);
	glRotatef(rot_x, 1.0f, 0.0f, 0.0f);
	glRotatef(rot_y, 0.0f, 1.0f, 0.0f);

	glutSolidCube(1);
	glPopMatrix();
	glutSwapBuffers();
}
void  mouseClick(int button , int state, int x, int y)
{  
	if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN) 
	{ 
		gButton = GLUT_LEFT_BUTTON;  
		//blahblah
		mouseX = x;
		mouseY = y;

		//if (m_select_face){
		//	gl_select(x, win_height - y);
		//}
	}

	if (button == GLUT_MIDDLE_BUTTON && state == GLUT_DOWN) {
		gButton = GLUT_MIDDLE_BUTTON;
		mouseX = x;
		mouseY = y;      	  

	}

	if (button == GLUT_RIGHT_BUTTON && state == GLUT_DOWN) {
		gButton = GLUT_RIGHT_BUTTON;
		mouseX = x;
		mouseY = y;
	}
	return ;
}

void mouseMove(int x, int y)
{
	Vector3D trans;

	/* rotation*/
	if (gButton == GLUT_LEFT_BUTTON ) 
	{
		////////////added////////////////
		rot_y += (x  - mouseX)/20.0;
		rot_x += (y - mouseY)/20.0;

		////////////////////////////////
		glutPostRedisplay();
	}

	/*xy translation */
	if (gButton == GLUT_MIDDLE_BUTTON) 
	{
		////////////added////////////////
		GLfloat scale_y = 10.f / win_height;
		GLfloat scale_x = 10.f / win_width;

		dx = scale_x * ( x - mouseX  );
		dy = scale_y * ( mouseY - y);
		////////////////////////////////
		//	  trans = Vector3D( scale_x * (  mouseX - x ), scale_y * ( y - mouseY ), 0 );
		//	  mouseX = x;
		//	  mouseY = y;
		//	  ObjTrans -= trans;
		glutPostRedisplay();
	}

	/* zoom in and out */
	if (gButton == GLUT_RIGHT_BUTTON ) 
	{
		double scale = 10./win_height;
		trans =  Vector3D(0,0, scale*(mouseY-y)   );
		mouseX = x;
		mouseY = y;
		ObjTrans = ObjTrans-trans;
		//or you can do :			  EyeTrans = EyeTrans+trans;
		glutPostRedisplay();
	}
}
/************************************************************************/
/*                                                                      */
void cvInit()
{
	if(!face_cascade.load(face_cascade_name)){
		printf("--(!)Error loading\n"); 
	};
	if(!eyes_cascade.load(eyes_cascade_name)){ 
		printf("--(!)Error loading\n"); 
	};

	capture = cvCaptureFromCAM( -1 );	
	cvSetCaptureProperty( capture, CV_CAP_PROP_FRAME_WIDTH, 256);
	cvSetCaptureProperty( capture, CV_CAP_PROP_FRAME_HEIGHT, 256);
}
void cvDetectFaceMotion()
{
	frame = cvQueryFrame( capture );

	if( !frame.empty() ){ 
		detectAndDisplay( frame ); 
	}
}
/************************************************************************/

//** @function main */
int main( int argc,  char** argv )
{
/************************************************************************/
/* opengl part                                                                     */

	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH );
	glutInitWindowPosition(100,100);

	int width = 640;
	int height = 480;
	whratio = (float)width /(float)height;
	glutInitWindowSize(width, height);
	glutCreateWindow("3D Scence Demo");
	glutDisplayFunc(renderScene);
	glutReshapeFunc(reshapeScene);
	glutMouseFunc(mouseClick);
	glutMotionFunc(mouseMove);
	//glutTimerFunc(100, captureCamImageCallback, val_time_callback);
	setupGLstate();
	
	cvInit();
	timer(0);
	glutMainLoop();
/************************************************************************/

/*
	CvCapture* capture;
	Mat frame;

	//-- 1. Load the cascades
	if( !face_cascade.load( face_cascade_name ) ){ printf("--(!)Error loading\n"); return -1; };
	if( !eyes_cascade.load( eyes_cascade_name ) ){ printf("--(!)Error loading\n"); return -1; };

	//-- 2. Read the video stream
	capture = cvCaptureFromCAM( -1 );	
	cvSetCaptureProperty( capture, CV_CAP_PROP_FRAME_WIDTH, 256);
	cvSetCaptureProperty( capture, CV_CAP_PROP_FRAME_HEIGHT, 256);
	
	if( capture )
	{
		while( true )
		{
			frame = cvQueryFrame( capture );

			//-- 3. Apply the classifier to the frame
			if( !frame.empty() )
			{ detectAndDisplay( frame ); }
			else
			{ printf(" --(!) No captured frame -- Break!"); break; }

			int c = waitKey(33);
			if( (char)c == 'c' ) { break; }
		}
	}*/
	return 0;
}

int getHorizontalMotion(Point p_end, Point p_beg)
{
	Point vec_motion = p_end - p_beg;
	//double sensity = vec_motion.y/vec_motion.x;
	//printf("Motion-X: %d, Motion-Y: %d \n", vec_motion.x, vec_motion.y);
//	if (abs(vec_motion.x) > 5){
//	std::cout<<vec_motion.x<<std::endl;
	if (abs(vec_motion.x)>5)
		rot_y += vec_motion.x*0.8;
	if (vec_motion.x > 0){
		//dx += 2;
		return M_LEFT;
	}else if (vec_motion.x < 0){
		//dx -= 2;
		return M_RIGHT;
	}
//	}
	return M_STATIC;
}
int getVerticalMotion(Point p_end, Point p_beg)
{
	Point vec_motion = p_end - p_beg;
	//double sensity = vec_motion.y/vec_motion.x;
	//printf("Motion-X: %d, Motion-Y: %d \n", vec_motion.x, vec_motion.y);
	if (abs(vec_motion.y) > 5)
		rot_x += -vec_motion.y*2;
	if (vec_motion.y > 0){
		//dy += 2;
		return M_DOWN;
	}else if (vec_motion.y < 0){
		//dy -= 2;
		return M_UP;
	}
//	}
	return M_STATIC;
}
//** @function detectAndDisplay */
int n_frame=0;
void detectAndDisplay( Mat frame )
{
	std::vector<Rect> faces;
	Mat frame_gray;

	cvtColor( frame, frame_gray, CV_BGR2GRAY );
	equalizeHist( frame_gray, frame_gray );

	//-- Detect faces
	face_cascade.detectMultiScale( frame_gray, faces, 1.1, 2, 0|CV_HAAR_SCALE_IMAGE, Size(30, 30) );

	for( int i = 0; i < faces.size(); i++ )
	{
		Point center( faces[i].x + faces[i].width*0.5, faces[i].y + faces[i].height*0.5 );
		//ellipse( frame, center, Size( faces[i].width*0.5, faces[i].height*0.5), 0, 0, 360, Scalar( 255, 0, 255 ), 4, 8, 0 );
		rectangle(frame, faces[i].tl(), faces[i].br(), Scalar(0, 255, 0), 3);

		Point bottom_left(faces[i].tl().x, faces[i].br().y);
		Point leftmost(center.x - faces[i].width, center.y);
		Point top_right(faces[i].br().x, faces[i].tl().y);
		Point upmost(center.x, center.y - faces[i].height);
		Point rightmost(center.x +  faces[i].width, center.y);
		Point downmost(center.x, center.y + faces[i].height);
		
		hor_move = M_STATIC;
		vert_move = M_STATIC;
		if (last_center.x!=0 && last_center.y!=0){
			hor_move = getHorizontalMotion(center, last_center);
			vert_move = getVerticalMotion(center, last_center);
		}
		if (hor_move != M_STATIC && vert_move != M_STATIC){
			if (hor_move==M_RIGHT){
				line(frame, faces[i].tl(), leftmost, Scalar(0, 0, 255), 3);	
				line(frame, bottom_left, leftmost, Scalar(0, 0, 255), 3);
			}else if(hor_move==M_LEFT){
				line(frame, top_right, rightmost, Scalar(0, 0, 255), 3);
				line(frame, faces[i].br(), rightmost, Scalar(0, 0, 255), 3);
			}

			if (vert_move==M_UP){
				line(frame, faces[i].tl(), upmost, Scalar(0, 0, 255), 3);
				line(frame, top_right, upmost, Scalar(0, 0, 255), 3);
			}else if (vert_move==M_DOWN){
				line(frame, faces[i].br(), downmost, Scalar(0, 0, 255), 3);
				line(frame, bottom_left, downmost, Scalar(0, 0, 255), 3);
			}
		}
		last_center = center;		
		
		Point rect = (faces[i].tl() -  faces[i].br());
		double face0_size = sqrt( (double)(rect.x*rect.x + rect.y*rect.y) );
//		printf("Face size: %f\n", face0_size);
//		printf("Face size ratio: %f\n", face0_size / last_box_size);
		if (i==0)
		{

			if ((fabs(face0_size / last_box_size-1)>0.05 &&fabs(face0_size / last_box_size-1)<0.1 ))
			{
				ObjTrans.z += (face0_size-last_box_size)/face0_size*10;

			}
			//		printf("Last Face size: %f\n", last_box_size);	
			last_box_size = face0_size;
		}
		
		++n_frame;

	}

	imshow( window_name, frame );
}